# coding: utf-8
import webbrowser
import pandas as pd
from pyecharts.charts import Page
from tqdm import tqdm
from typing import List, Callable
from ..analyze import CZSC, RawBar
from ..utils.echarts_plot import heat_map
from ..utils.kline_generator import KlineGeneratorBy1Min, bar_end_time
from ..factors import CzscFactors


def cal_nbar_percentile(k: dict, kn: List[dict], n: int) -> float:
    """计算 N 周期区间百分位

    假设选股策略确认选股时间为 T，股价为 P，T+N 区间内的K线最高价为 MAX_P，最低价为 MIN_P，则
    N周期区间百分位 = (MAX_P - P) / (MAX_P - MIN_P) * 100;
    N周期区间百分位值域为(0, 100)，值越大，说明离区间最大值越远，对于做多交易而言有更高的安全性。

    :param k: dict
        信号出现时的 K 线，如：
        {'symbol': '000001.SH',
         'dt': '2007-04-05 15:00:00',
         'open': 3286.16,
         'close': 3319.14,
         'high': 3326.92,
         'low': 3259.63,
         'vol': 114528051.0}

    :param kn: List[dict]
        信号出现后的 N 根 K 线，如：
        [{'symbol': '000001.SH',
          'dt': '2007-04-06 15:00:00',
          'open': 3287.68,
          'close': 3323.59,
          'high': 3334.22,
          'low': 3273.86,
          'vol': 119644881.0},
         {'symbol': '000001.SH',
          'dt': '2007-04-09 15:00:00',
          'open': 3333.42,
          'close': 3398.95,
          'high': 3399.51,
          'low': 3333.26,
          'vol': 137314104.0}]

    :param n: int
        周期数 N

    :return: float
    """
    assert len(kn) == n, "计算 {} 周期区间百分位时传入的 kn 数量为 {}".format(n, len(kn))
    c = k['close']
    min_p = min([x['low'] for x in kn] + [c])
    max_p = max([x['high'] for x in kn] + [c])

    if max_p == min_p:
        return 0
    else:
        percentile = round((max_p - c) / (max_p - min_p) * 100, 2)
        return percentile


def cal_nbar_income(k: dict, kn: List[dict], n: int) -> float:
    """计算 N 周期区间收益

    假设选股策略确认选股时间为 T，股价为 P，第 T+N 根K线收盘价为 C，则
    N周期绝对收益 = (C - P) / P * 10000;
    N周期绝对收益值域为(-10000, +inf)，值越大，说明绝对收益越大，对于做多交易而言有更高的收益率。

    :param k: dict
        信号出现时的 K 线，如：
        {'symbol': '000001.SH',
         'dt': '2007-04-05 15:00:00',
         'open': 3286.16,
         'close': 3319.14,
         'high': 3326.92,
         'low': 3259.63,
         'vol': 114528051.0}

    :param kn: List[dict]
        信号出现后的 N 根 K 线，如：
        [{'symbol': '000001.SH',
          'dt': '2007-04-06 15:00:00',
          'open': 3287.68,
          'close': 3323.59,
          'high': 3334.22,
          'low': 3273.86,
          'vol': 119644881.0},
         {'symbol': '000001.SH',
          'dt': '2007-04-09 15:00:00',
          'open': 3333.42,
          'close': 3398.95,
          'high': 3399.51,
          'low': 3333.26,
          'vol': 137314104.0}]

    :param n: int
        周期数 N

    :return: float
    """
    assert len(kn) == n, "计算 {} 周期区间收益时传入的 kn 数量为 {}".format(n, len(kn))
    c = k['close']
    last_c = kn[-1]['close']

    income = round((last_c - c) / c * 10000, 2)
    return income


class FactorsResearcher:
    """因子研究员"""

    def __init__(self, factors, n_list=None):
        self.factors = factors
        self.performance = None
        if not n_list:
            self.n_list = (8, 13, 21, 34, 55, 89, 144, 233, 377, 610)
        else:
            self.n_list = n_list

        # 使用 kg 合成各级别K线
        self.kg = KlineGeneratorBy1Min()
        for row in factors:
            self.kg.update(RawBar(symbol=row['symbol'], dt=pd.to_datetime(row['dt']), open=row['open'],
                                  close=row['close'], high=row['high'], low=row['low'], vol=row['vol']))
        self.calculate_nbar()

    def remove_duplicates(self, name, replace_value=False):
        """清理指定因子的连续信号

        :param replace_value:
        :param name: 因子名
        :return:
        """
        for i in range(len(self.factors)):
            if i < 5:
                self.factors[i]['{}_duplicated'.format(name)] = replace_value
            else:
                rows = self.factors[i-4: i]
                if rows[-1][name] in [x[name] for x in rows[:-1]]:
                    self.factors[i]['{}_duplicated'.format(name)] = replace_value
                else:
                    self.factors[i]['{}_duplicated'.format(name)] = rows[-1][name]

    def show_klines(self, file_html, open_in_browser=True):
        kf = CzscFactors(self.kg, max_count=100000000)
        kf.take_snapshot(file_html)
        if open_in_browser:
            webbrowser.open(file_html)

    def calculate_nbar(self):
        """计算 nbar 相关特征"""
        max_len = len(self.factors)
        for i, k in tqdm(enumerate(self.factors), desc="calculate_nbar"):
            k['base'] = 1
            for n in self.n_list:
                if i + n + 1 < max_len:
                    kn = self.factors[i + 1: i + n + 1]
                    k['n{}b_均收益'.format(n)] = cal_nbar_income(k, kn, n)
                    k['n{}b_百分位'.format(n)] = cal_nbar_percentile(k, kn, n)
                else:
                    k['n{}b_均收益'.format(n)] = 0
                    k['n{}b_百分位'.format(n)] = 0

    def factors_performance(self, file_xlsx, factors_col):
        """查看所有因子的表现"""
        esp = 0.00001
        if 'base' not in factors_col:
            factors_col = ['base'] + factors_col

        results = []
        for factor_name in tqdm(factors_col, desc="factors_performance"):
            groups = dict()
            for row in self.factors:
                v = groups.get(row[factor_name], [])
                v.append(row)
                groups[row[factor_name]] = v

            for v, samples in groups.items():
                res = {"factor_name": factor_name, "factor_value": v,
                       "count": len(samples), "pct": (len(samples) / len(self.factors)) * 100}
                for n in self.n_list:
                    k1 = 'n{}b_均收益'.format(n)
                    res[k1] = sum([x[k1] for x in samples]) / (len(samples) + esp)
                    k2 = 'n{}b_百分位'.format(n)
                    res[k2] = sum([x[k2] for x in samples]) / (len(samples) + esp)
                    s1 = [x for x in samples if x[k1] > 0]
                    s2 = [x for x in samples if x[k1] < 0]
                    res[k1.replace("均收益", "胜率")] = (len(s1) / (len(samples) + esp)) * 100
                    s1_mean = sum([x[k1] for x in s1]) / (len(s1) + esp)
                    s2_mean = sum([abs(x[k1]) for x in s2]) / (len(s2) + esp)
                    res[k1.replace("均收益", "盈亏比")] = s1_mean / (s2_mean + esp)
                results.append(res)

        df = pd.DataFrame(results)
        cols = ['factor_name', 'factor_value', 'count', 'pct']
        for n in self.n_list:
            cols.append('n{}b_均收益'.format(n))
            cols.append('n{}b_百分位'.format(n))
            cols.append('n{}b_胜率'.format(n))
            cols.append('n{}b_盈亏比'.format(n))

        for col in cols[3:]:
            df[col] = df[col].apply(lambda x: round(x, 2))
        df[cols].to_excel(file_xlsx, index=False)
        print("factors performance saved into {}".format(file_xlsx))
        self.performance = df[cols]

    def show_bs(self, bs, freq="30分钟", file_html=None, open_in_browser=True, width="1400px", height='580px'):
        """在K线图上将 factor 标记出来

        :param height:
        :param width:
        :param open_in_browser:
        :param file_html:
        :param bs: list of dict
            [{"dt": dt, "mark": buy or sell, "price": 0}]
        :param freq: str
            指定K线图级别
        :return:
        """
        kline = self.kg.get_kline(freq, count=100000000)
        ka = CZSC(kline, max_count=100000000)
        chart = ka.to_echarts(width=width, height=height)
        if file_html:
            chart.render(file_html)
            if open_in_browser:
                webbrowser.open(file_html)
        return chart

    def show_factor(self, name, value, file_html, freq="30分钟", open_in_browser=True):
        """查看因子某个值的结果

        :param freq:
        :param name: str
            因子名
        :param value: any
            因子值
        :param file_html: str
        :param open_in_browser: bool
        :return:
        """
        freq_m = {"1分钟": 1, "5分钟": 5, "30分钟": 30}
        rows = [x for x in self.factors if x[name] == value]
        value1 = [{"x": s['dt'], "y": k, "heat": v} for s in rows for k, v in s.items() if "百分位" in k]
        value2 = [{"x": s['dt'], "y": k, "heat": v} for s in rows for k, v in s.items() if "均收益" in k]
        x_label = [s['dt'] for s in rows]
        y_label1 = [k for k, _ in rows[0].items() if "百分位" in k]
        y_label2 = [k for k, _ in rows[0].items() if "均收益" in k]

        hm1 = heat_map(value1, x_label=x_label, y_label=y_label1,
                       title="{}历史表现评估（N周期区间百分位）".format(name), width="1400px", height="480px")
        hm2 = heat_map(value2, x_label=x_label, y_label=y_label2,
                       title="{}历史表现评估（N周期区间收益）".format(name), width="1400px", height="480px")

        bs = [{"dt": bar_end_time(s['dt'], m=freq_m[freq]), "mark": "buy", "price": round(s['close'], 2)} for s in rows]
        chart_kline = self.show_bs(bs, freq=freq, width="1400px", height='480px')
        page = Page(layout=Page.DraggablePageLayout, page_title="{}".format(name))
        page.add(hm1, hm2, chart_kline)

        if file_html:
            page.render(file_html)
            if open_in_browser:
                webbrowser.open(file_html)
        return page

    def add_temp_factors(self, functions: List[Callable]) -> None:
        """新增因子，用于快速研究

        :param functions:
        :return: None
        """
        for factor in functions:
            name = factor.__doc__
            for row in tqdm(self.factors, desc="add {}".format(name)):
                row[name] = factor(row)





# coding: utf-8
import pandas as pd
from datetime import datetime
from .factors import KlineGeneratorBy1Min, CzscFactors
from .data.jq import get_kline
from .enum import FdNine, FdFive, FdSeven, FdThree, Factors

class CzscTrader:
    """缠中说禅股票 选股/择时"""
    def __init__(self, symbol, max_count=1000, end_date=None):
        """
        :param symbol:
        """
        self.symbol = symbol
        if end_date:
            self.end_date = pd.to_datetime(end_date)
        else:
            self.end_date = datetime.now()
        self.max_count = max_count
        self.__generate_factors()
        self.freqs = ['1分钟', '5分钟', '30分钟', '日线']

    def __generate_factors(self):
        symbol = self.symbol
        max_count = self.max_count
        end_date = self.end_date

        kg = KlineGeneratorBy1Min(max_count=max_count*2, freqs=['1分钟', '5分钟', '15分钟', '30分钟', '日线'])
        k1min = get_kline(symbol, end_date=end_date, freq="1min", count=max_count)
        k5min = get_kline(symbol, end_date=end_date, freq="5min", count=max_count)
        k15min = get_kline(symbol, end_date=end_date, freq="15min", count=max_count)
        k30min = get_kline(symbol, end_date=end_date, freq="30min", count=max_count)
        kd = get_kline(symbol, end_date=end_date, freq="D", count=max_count)
        kg.init_kline("1分钟", k1min)
        kg.init_kline("5分钟", k5min)
        kg.init_kline("15分钟", k15min)
        kg.init_kline("30分钟", k30min)
        kg.init_kline("日线", kd)
        kf = CzscFactors(kg, max_count=max_count)
        self.kf = kf
        self.s = kf.s
        self.end_dt = self.kf.end_dt

    def run_selector(self):
        """执行选股：优先输出大级别的机会"""
        s = self.s
        if s['日线右侧多头因子'] in [Factors.DLA1.value, Factors.DLA2.value, Factors.DLA3.value, Factors.DLA4.value]:
            return s['日线右侧多头因子']

        ka = self.kf.kas['30分钟']
        max_high = max([x.high for x in ka.bi_list[-10:]])

        # third_bs = ["三买A1", "三买B1", "三买C1", "三买D1"]
        if "三买" in s['30分钟_第N笔的五笔形态']:
            if s['1分钟_第N笔的七笔形态'] == FdSeven.L3A1.value:
                return "30分钟第三买点且BaA式右侧底A"
            elif max_high == ka.bi_list[-2].high:
                return "30分钟第三买点且第4笔创近9笔新高"
            else:
                return "30分钟第三买点"

        # nine_values = [x.value for x in FdNine.__members__.values() if x.name[0] in ["L", "S"]]
        # seven_values = [x.value for x in FdSeven.__members__.values() if x.name[0] in ["L", "S"]]
        # if s['30分钟_第N笔的七笔形态'] in seven_values:
        #     return "30分钟_第N笔的七笔形态_{}".format(s['30分钟_第N笔的七笔形态'])
        # if s['30分钟_第N笔的九笔形态'] in nine_values:
        #     return "30分钟_第N笔的九笔形态_{}".format(s['30分钟_第N笔的九笔形态'])
        return "other"

    def run_history(self):
        """对已经完成的三买走势进行研究"""
        s = self.s
        if "三买" in s['30分钟_第N-2笔的五笔形态']:
            return "30分钟第N-2笔第三买点"
        return "other"

    def take_snapshot(self, file_html, width="1400px", height="680px"):
        self.kf.take_snapshot(file_html, width, height)



# coding: utf-8

from enum import Enum

class Mark(Enum):
    D = "底分型"
    G = "顶分型"

class Direction(Enum):
    Up = "向上"
    Down = "向下"

class Freq(Enum):
    F1 = "1分钟"
    F5 = "5分钟"
    F15 = "15分钟"
    F30 = "30分钟"
    F60 = "60分钟"
    D = "日线"
    W = "周线"
    M = "月线"

class FdThree(Enum):
    """三段形态分类"""
    # 多头形态
    L1 = "向下不重合"
    L2 = "向下奔走型中枢"
    L3 = "向下三角收敛中枢"
    L4 = "向下三角扩张中枢"
    L5 = "向下盘背中枢"
    L6 = "向下无背中枢"

    # 空头形态
    S1 = "向上不重合"
    S2 = "向上奔走型中枢"
    S3 = "向上三角收敛中枢"
    S4 = "向上三角扩张中枢"
    S5 = "向上盘背中枢"
    S6 = "向上无背中枢"

    Other = "其他"


class FdFive(Enum):
    """五段形态分类"""
    L1A1 = "向下三角收敛中枢"
    L1B1 = "向下三角扩张中枢"

    L2A1 = "aAb式底背驰A"
    L2B1 = "aAb式底背驰B"
    L2C1 = "aAb式底背驰C"

    L3A1 = "类趋势底背驰"

    L4A1 = "三买A1"
    L4A2 = "三买A2"
    L4B1 = "三买B1"
    L4B2 = "三买B2"
    L4C1 = "三买C1"
    L4C2 = "三买C2"
    L4D1 = "三买D1"
    L4D2 = "三买D2"

    L5A1 = "cA底部A"
    L5B1 = "cA底部B"

    S1A1 = "向上三角收敛中枢"
    S1B1 = "向上三角扩张中枢"

    S2A1 = "aAb式顶背驰A"
    S2B1 = "aAb式顶背驰B"
    S2C1 = "aAb式顶背驰C"

    S3A1 = "类趋势顶背驰"

    S4A1 = "三卖A1"
    S4A2 = "三卖A2"
    S4B1 = "三卖B1"
    S4B2 = "三卖B2"
    S4C1 = "三卖C1"
    S4C2 = "三卖C2"
    S4D1 = "三卖D1"
    S4D2 = "三卖D2"

    S5A1 = "cA顶部A"
    S5B1 = "cA顶部B"

    Other = "其他"

class FdSeven(Enum):
    """七段形态分类"""
    L1A1 = "aAb式底背驰"
    L2A1 = "aAbcd式底背驰"
    L3A1 = "BaA式右侧底A"
    L3B1 = "BaA式右侧底B"
    L4A1 = "类趋势底背驰"

    S1A1 = "aAb式顶背驰"
    S2A1 = "aAbcd式顶背驰"
    S3A1 = "BaA式右侧顶A"
    S3B1 = "BaA式右侧顶B"
    S4A1 = "类趋势顶背驰"

    Other = "其他"

class FdNine(Enum):
    """九段形态分类"""
    L1A1 = "类趋势底背驰"

    L2A1 = "含有一个上下上方向中枢A"
    L2B1 = "含有一个上下上方向中枢B"
    L2C1 = "含有一个上下上方向中枢C"

    L3A1 = "标准aAbBc向下趋势"

    L4A1 = "上中枢七段下跌"
    L4B1 = "下中枢七段下跌"

    # =================================

    S1A1 = "类趋势顶背驰"

    S2A1 = "含有一个下上下方向中枢A"
    S2B1 = "含有一个下上下方向中枢B"
    S2C1 = "含有一个下上下方向中枢C"

    S3A1 = "标准aAbBc向上趋势"

    S4A1 = "上中枢七段上涨"
    S4B1 = "下中枢七段上涨"

    Other = "其他"

class Factors(Enum):
    DLA1 = "日线右侧多头1"   # 日线第N笔向下结束，且30分钟第N笔的七笔形态为BaA式右侧底B
    DLA2 = "日线右侧多头2"   # 日线向下笔以超强底分结束，且5分钟第N笔的七笔形态为BaA式右侧底A
    DLA3 = "日线右侧多头3"   # 日线第N笔的五笔形态为底背驰，且30分钟第N笔的七笔形态为BaA式右侧底B
    DLA4 = "日线右侧多头4"   # 日线第N笔的七笔形态为底背驰，且30分钟第N笔的七笔形态为BaA式右侧底B

    DLB1 = "日线左侧多头1"   # 日线第N笔向下延伸，且30分钟第N笔的七笔形态为aAbcd式底背驰或类趋势底背驰

    F30LA1 = "30分钟右侧多头1"   # 5分钟第N笔的五笔形态为三买且1分钟第N笔的七笔形态为BaA式右侧底B

    F5LA1 = "5分钟右侧多头1"   # 5分钟第N笔向下结束，且1分钟第N笔的七笔形态为BaA式右侧底B
    F5LA2 = "5分钟右侧多头2"   # 5分钟第N笔向下结束，且1分钟第N笔的七笔形态为BaA式右侧底B
    F5LA3 = "5分钟右侧多头3"   # 5分钟第N笔的七笔形态为底背驰，且1分钟第N笔的七笔形态为BaA式右侧底B

    F5SA1 = "5分钟右侧空头1"   # 5分钟第N笔向上结束，且1分钟第N笔的七笔形态为BaA式右侧顶B
    F5SA2 = "5分钟右侧空头2"   # 5分钟第N笔的五笔形态为顶背驰，且1分钟第N笔的七笔形态为BaA式右侧顶B
    F5SA3 = "5分钟右侧空头3"   # 5分钟第N笔的七笔形态为顶背驰，且1分钟第N笔的七笔形态为BaA式右侧顶B

    # =================
    DSA1 = "日线右侧空头1"

    DSB1 = "日线左侧空头1"    # 日线第N笔方向向上且没有发生向上破坏，且30分钟第N笔的五笔形态为顶背驰
    DSB2 = "日线左侧空头2"    # 日线第N笔方向向上且没有发生向上破坏，且30分钟第N笔的七笔形态为顶背驰

    Other = "其他"




# coding: utf-8
import sys
import warnings

sys.path.insert(0, '.')
sys.path.insert(0, '..')
import os
import pandas as pd
import czsc
from czsc.cobra.analyst import cal_nbar_income, cal_nbar_percentile

warnings.warn("czsc version is {}".format(czsc.__version__))

cur_path = os.path.split(os.path.realpath(__file__))[0]


def test_nbar():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    bars = kline.to_dict("records")

    i = 100
    n = 10
    k1 = bars[i]
    kn = bars[i+1: i+n+1]

    ni = cal_nbar_income(k1, kn, n)
    assert ni == 391.31

    np = cal_nbar_percentile(k1, kn, n)
    assert np == 87.06







# coding: utf-8

from typing import List
from collections import OrderedDict
from datetime import datetime
import pandas as pd
from .objects import Mark, Direction, BI, FX, RawBar, NewBar
from .utils.echarts_plot import kline_pro
from .signals import check_three_fd, check_five_fd, check_seven_fd, check_nine_fd

def remove_include(k1: NewBar, k2: NewBar, k3: RawBar):
    """去除包含关系：输入三根k线，其中k1和k2为没有包含关系的K线，k3为原始K线"""
    if k1.high < k2.high:
        direction = Direction.Up
    elif k1.high > k2.high:
        direction = Direction.Down
    else:
        direction = Direction.Up

    # 判断 k2 和 k3 之间是否存在包含关系，有则处理
    if (k2.high <= k3.high and k2.low >= k3.low) or (k2.high >= k3.high and k2.low <= k3.low):
        if direction == Direction.Up:
            high = max(k2.high, k3.high)
            low = max(k2.low, k3.low)
            dt = k2.dt if k2.high >= k3.high else k3.dt
        elif direction == Direction.Down:
            high = min(k2.high, k3.high)
            low = min(k2.low, k3.low)
            dt = k2.dt if k2.low <= k3.low else k3.dt
        else:
            raise ValueError

        if k3.open > k3.close:
            open_ = high
            close = low
        else:
            open_ = low
            close = high
        vol = k2.vol + k3.vol
        k4 = NewBar(symbol=k3.symbol, dt=dt, open=open_,
                    close=close, high=high, low=low, vol=vol, elements=k2.elements + [k3])
        return True, k4
    else:
        k4 = NewBar(symbol=k3.symbol, dt=k3.dt, open=k3.open,
                    close=k3.close, high=k3.high, low=k3.low, vol=k3.vol, elements=[k3])
        return False, k4


def check_fx(k1: NewBar, k2: NewBar, k3: NewBar):
    """查找分型"""
    fx = None
    if k1.high < k2.high > k3.high:
        power = "强" if k3.close < k1.low else "弱"
        fx = FX(symbol=k1.symbol, dt=k2.dt, mark=Mark.G, high=k2.high, low=k2.low,
                fx=k2.high, elements=[k1, k2, k3], power=power)
    if k1.low > k2.low < k3.low:
        power = "强" if k3.close > k1.high else "弱"
        fx = FX(symbol=k1.symbol, dt=k2.dt, mark=Mark.D, high=k2.high, low=k2.low,
                fx=k2.low, elements=[k1, k2, k3], power=power)
    return fx


def check_bi(bars: List[NewBar]):
    """输入一串无包含关系K线，查找其中的一笔"""
    fxs = []
    for i in range(1, len(bars)-1):
        fx = check_fx(bars[i-1], bars[i], bars[i+1])
        if isinstance(fx, FX):
            fxs.append(fx)

    if len(fxs) < 2:
        return None, bars

    fx_a = fxs[0]
    try:
        if fxs[0].mark == Mark.D:
            direction = Direction.Up
            fxs_b = [x for x in fxs if x.mark == Mark.G]
            fx_b = fxs_b[0]
            for fx in fxs_b:
                if fx.high > fx_b.high:
                    fx_b = fx
            high = fx_b.high
            low = fx_a.low
        elif fxs[0].mark == Mark.G:
            direction = Direction.Down
            fxs_b = [x for x in fxs if x.mark == Mark.D]
            fx_b = fxs_b[0]
            for fx in fxs_b[1:]:
                if fx.low < fx_b.low:
                    fx_b = fx
            high = fx_a.high
            low = fx_b.low
        else:
            raise ValueError
    except:
        return None, bars

    bars_a = [x for x in bars if x.dt <= fx_b.elements[2].dt]
    bars_b = [x for x in bars if x.dt >= fx_b.elements[0].dt]
    max_high_b = max([x.high for x in bars_b])
    min_low_b = min([x.low for x in bars_b])
    if (direction == Direction.Up and max_high_b > fx_b.high) \
            or (direction == Direction.Down and min_low_b < fx_b.low):
        return None, bars

    ab_include = (fx_a.high > fx_b.high and fx_a.low < fx_b.low) or (fx_a.high < fx_b.high and fx_a.low > fx_b.low)
    if len(bars_a) >= 7 and not ab_include:
        power_price = abs(fx_b.fx - fx_a.fx)
        bi = BI(symbol=fx_a.symbol, fx_a=fx_a, fx_b=fx_b, direction=direction,
                power=power_price, high=high, low=low, bars=bars_a)
        return bi, bars_b
    else:
        return None, bars


def get_sub_span(bis: List[BI], start_dt: [datetime, str], end_dt: [datetime, str]) -> List[BI]:
    """获取子区间（这是进行多级别联立分析的关键步骤）

    :param bis: 笔的列表
    :param start_dt: 子区间开始时间
    :param end_dt: 子区间结束时间
    :return: 子区间
    """
    start_dt = pd.to_datetime(start_dt)
    end_dt = pd.to_datetime(end_dt)
    sub = []
    for bi in bis:
        if bi.fx_b.dt > start_dt > bi.fx_a.dt:
            sub.append(bi)
        elif start_dt <= bi.fx_a.dt < bi.fx_b.dt <= end_dt:
            sub.append(bi)
        elif bi.fx_a.dt < end_dt < bi.fx_b.dt:
            sub.append(bi)
        else:
            continue
    return sub


class CZSC:
    def __init__(self, bars: List[RawBar], freq: str, max_count=1000):
        """

        :param bars: K线数据
        :param max_count: int
            最大保存的K线数量
        """
        self.max_count = max_count
        self.bars_raw = []  # 原始K线序列
        self.bars_ubi = []  # 未完成笔的无包含K线序列
        self.bi_list: List[BI] = []
        self.symbol = bars[0].symbol
        self.freq = freq

        for bar in bars:
            self.update(bar)

    def __update_bi(self):
        bars_ubi = self.bars_ubi

        # 查找笔
        if not self.bi_list:
            bi, bars_ubi_ = check_bi(bars_ubi)
            if isinstance(bi, BI):
                self.bi_list.append(bi)
            self.bars_ubi = bars_ubi_
            return

        last_bi = self.bi_list[-1]

        # 如果上一笔被破坏，将上一笔的bars与bars_ubi进行合并
        min_low_ubi = min([x.low for x in bars_ubi])
        max_high_ubi = max([x.high for x in bars_ubi])

        if last_bi.direction == Direction.Up and max_high_ubi > last_bi.high:
            if min_low_ubi < last_bi.low and len(self.bi_list) > 2:
                bars_ubi_a = self.bi_list[-2].bars \
                             + [x for x in self.bi_list[-1].bars if x.dt > self.bi_list[-2].bars[-1].dt] \
                             + [x for x in bars_ubi if x.dt > self.bi_list[-1].bars[-1].dt]
                self.bi_list.pop(-1)
                self.bi_list.pop(-1)
            else:
                bars_ubi_a = last_bi.bars + [x for x in bars_ubi if x.dt > last_bi.bars[-1].dt]
                self.bi_list.pop(-1)
        elif last_bi.direction == Direction.Down and min_low_ubi < last_bi.low:
            if max_high_ubi > last_bi.high and len(self.bi_list) > 2:
                bars_ubi_a = self.bi_list[-2].bars \
                             + [x for x in self.bi_list[-1].bars if x.dt > self.bi_list[-2].bars[-1].dt] \
                             + [x for x in bars_ubi if x.dt > self.bi_list[-1].bars[-1].dt]
                self.bi_list.pop(-1)
                self.bi_list.pop(-1)
            else:
                bars_ubi_a = last_bi.bars + [x for x in bars_ubi if x.dt > last_bi.bars[-1].dt]
                self.bi_list.pop(-1)
        else:
            bars_ubi_a = bars_ubi

        if len(bars_ubi_a) > 300:
            print("{} - {} 未完成笔延伸超长，延伸数量: {}".format(self.symbol, self.freq, len(bars_ubi_a)))
        bi, bars_ubi_ = check_bi(bars_ubi_a)
        self.bars_ubi = bars_ubi_
        if isinstance(bi, BI):
            self.bi_list.append(bi)

    def get_signals(self):
        s = OrderedDict({"symbol": self.symbol, "dt": self.bars_raw[-1].dt, "close": self.bars_raw[-1].close})
        s.update({
            "最近三根无包含K线形态": "other",
            "未完成笔的延伸长度": 0,

            "第N笔方向": "other",
            "第N笔结束标记的上边沿": 0,
            "第N笔结束标记的下边沿": 0,
            "第N笔结束标记的分型强弱": 0,

            "第N-1笔结束标记的上边沿": 0,
            "第N-1笔结束标记的下边沿": 0,
            "第N-1笔结束标记的分型强弱": 0,

            "第N笔的三笔形态": "other",
            "第N-1笔的三笔形态": "other",
            "第N-2笔的三笔形态": "other",
            "第N-3笔的三笔形态": "other",

            "第N笔的五笔形态": "other",
            "第N-1笔的五笔形态": "other",
            "第N-2笔的五笔形态": "other",
            "第N-3笔的五笔形态": "other",

            "第N笔的七笔形态": "other",
            "第N-1笔的七笔形态": "other",
            "第N-2笔的七笔形态": "other",
            "第N-3笔的七笔形态": "other",

            "第N笔的九笔形态": "other",
            "第N-1笔的九笔形态": "other",
            "第N-2笔的九笔形态": "other",
            "第N-3笔的九笔形态": "other",
        })
        s['未完成笔的延伸长度'] = len(self.bars_ubi)

        if s['未完成笔的延伸长度'] > 3:
            k1, k2, k3 = self.bars_ubi[-3:]
            tri = check_fx(k1, k2, k3)
            if isinstance(tri, FX):
                s['最近三根无包含K线形态'] = tri.power + tri.mark.value

        bis = self.bi_list
        if len(bis) > 3:
            direction = bis[-1].direction
            s['第N笔方向'] = direction.value
            s['第N笔结束标记的上边沿'] = bis[-1].fx_b.high
            s['第N笔结束标记的下边沿'] = bis[-1].fx_b.low
            s['第N笔结束标记的分型强弱'] = bis[-1].fx_b.power

            s['第N-1笔结束标记的上边沿'] = bis[-2].fx_b.high
            s['第N-1笔结束标记的下边沿'] = bis[-2].fx_b.low
            s['第N-1笔结束标记的分型强弱'] = bis[-2].fx_b.power

        if len(self.bi_list) > 8:
            bis = self.bi_list
            s['第N笔的三笔形态'] = check_three_fd(bis[-3:])
            s['第N-1笔的三笔形态'] = check_three_fd(bis[-4:-1])
            s['第N-2笔的三笔形态'] = check_three_fd(bis[-5:-2])
            s['第N-3笔的三笔形态'] = check_three_fd(bis[-6:-3])

            s['第N笔的五笔形态'] = check_five_fd(bis[-5:])
            s['第N-1笔的五笔形态'] = check_five_fd(bis[-6:-1])

        if len(self.bi_list) > 10:
            bis = self.bi_list
            s['第N-2笔的五笔形态'] = check_five_fd(bis[-7:-2])
            s['第N-3笔的五笔形态'] = check_five_fd(bis[-8:-3])

            s['第N笔的七笔形态'] = check_seven_fd(bis[-7:])
            s['第N-1笔的七笔形态'] = check_seven_fd(bis[-8:-1])

        if len(self.bi_list) > 12:
            bis = self.bi_list
            s['第N-2笔的七笔形态'] = check_seven_fd(bis[-9:-2])
            s['第N-3笔的七笔形态'] = check_seven_fd(bis[-10:-3])

            s['第N笔的九笔形态'] = check_nine_fd(bis[-9:])
            s['第N-1笔的九笔形态'] = check_nine_fd(bis[-10:-1])

        if len(self.bi_list) > 15:
            bis = self.bi_list
            s['第N-2笔的九笔形态'] = check_nine_fd(bis[-11:-2])
            s['第N-3笔的九笔形态'] = check_nine_fd(bis[-12:-3])

        return {"{}_{}".format(self.freq, k) if k not in ['symbol', 'dt', 'close'] else k: v for k, v in s.items()}

    def update(self, bar: RawBar):
        """更新分析结果

        :param bar: 单根K线对象
        """
        # 更新K线序列
        if not self.bars_raw or bar.dt != self.bars_raw[-1].dt:
            self.bars_raw.append(bar)
            last_bars = [bar]
        else:
            self.bars_raw[-1] = bar
            last_bars = self.bars_ubi[-1].elements
            last_bars[-1] = bar
            self.bars_ubi.pop(-1)

        # 去除包含关系
        bars_ubi = self.bars_ubi
        for bar in last_bars:
            if len(bars_ubi) < 2:
                bars_ubi.append(NewBar(symbol=bar.symbol, dt=bar.dt, open=bar.open, close=bar.close,
                                       high=bar.high, low=bar.low, vol=bar.vol, elements=[bar]))
            else:
                k1, k2 = bars_ubi[-2:]
                has_include, k3 = remove_include(k1, k2, bar)
                if has_include:
                    bars_ubi[-1] = k3
                else:
                    bars_ubi.append(k3)

        self.bars_ubi = bars_ubi

        # 更新 笔
        self.__update_bi()
        self.bars_raw = self.bars_raw[-self.max_count:]
        self.bi_list = self.bi_list[-(self.max_count // 7):]

    def to_echarts(self, width: str = "1400px", height: str = '580px'):
        kline = [x.__dict__ for x in self.bars_raw]
        if len(self.bi_list) > 0:
            bi = [{'dt': x.fx_a.dt, "bi": x.fx_a.fx} for x in self.bi_list] + \
                 [{'dt': self.bi_list[-1].fx_b.dt, "bi": self.bi_list[-1].fx_b.fx}]
        else:
            bi = None
        chart = kline_pro(kline, bi=bi, width=width, height=height, title="{}-{}".format(self.symbol, self.freq))
        return chart



# coding: utf-8
from collections import OrderedDict
from pyecharts.charts import Tab
from pyecharts.components import Table
from pyecharts.options import ComponentTitleOpts
from typing import List
from .analyze import CZSC
from .utils.kline_generator import KlineGeneratorBy1Min, KlineGeneratorByTick
from .objects import RawBar
from .enum import Factors, FdNine, FdFive, FdSeven, FdThree, Direction


class CzscFactors:
    """缠中说禅技术分析理论之多级别联立因子"""
    def __init__(self, kg: [KlineGeneratorByTick, KlineGeneratorBy1Min], max_count: int = 1000):
        """

        :param kg: 基于tick或1分钟的K线合成器
        :param max_count: 单个级别最大K线数量
        """
        assert max_count >= 1000, "为了保证因子能够顺利计算，max_count 不允许设置小于1000"
        self.kg = kg
        self.freqs = kg.freqs
        klines = self.kg.get_klines({k: max_count for k in self.freqs})
        self.kas = {k: CZSC(klines[k], freq=k, max_count=max_count) for k in klines.keys()}
        self.symbol = self.kas["1分钟"].symbol
        self.end_dt = self.kas["1分钟"].bars_raw[-1].dt
        self.latest_price = self.kas["1分钟"].bars_raw[-1].close
        self.s = self._calculate_factors()
        self.cache = OrderedDict()

    def take_snapshot(self, file_html=None, width="1400px", height="580px"):
        """获取快照

        :param file_html: str
            交易快照保存的 html 文件名
        :param width: str
            图表宽度
        :param height: str
            图表高度
        :return:
        """
        tab = Tab(page_title="{}@{}".format(self.symbol, self.end_dt.strftime("%Y-%m-%d %H:%M")))
        for freq in self.freqs:
            chart = self.kas[freq].to_echarts(width, height)
            tab.add(chart, freq)

        t1 = Table()
        t1.add(["名称", "数据"], [[k, v] for k, v in self.s.items() if "_" in k])
        t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
        tab.add(t1, "信号表")

        t2 = Table()
        t2.add(["名称", "数据"], [[k, v] for k, v in self.s.items() if "_" not in k])
        t2.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅因子表", subtitle=""))
        tab.add(t2, "因子表")

        if file_html:
            tab.render(file_html)
        else:
            return tab

    def _calculate_signals(self):
        """计算信号"""
        s = OrderedDict()
        for freq, ks in self.kas.items():
            s.update(ks.get_signals())

        s.update(self.kas['1分钟'].bars_raw[-1].__dict__)
        return s

    def _calculate_factors(self):
        """计算因子"""
        s = self._calculate_signals()

        s.update({
            "日线右侧多头因子": Factors.Other.value,
            "日线左侧多头因子": Factors.Other.value,

            "日线右侧空头因子": Factors.Other.value,
            "日线左侧空头因子": Factors.Other.value,

            "30分钟右侧多头因子": Factors.Other.value,
            "30分钟左侧多头因子": Factors.Other.value,

            "30分钟右侧空头因子": Factors.Other.value,
            "30分钟左侧空头因子": Factors.Other.value,

            "5分钟右侧多头因子": Factors.Other.value,
            "5分钟左侧多头因子": Factors.Other.value,

            "5分钟右侧空头因子": Factors.Other.value,
            "5分钟左侧空头因子": Factors.Other.value,
        })

        if "日线" in self.freqs and "30分钟" in self.freqs and "5分钟" in self.freqs:
            if s['日线_第N笔方向'] == Direction.Down.value and s['30分钟_第N笔的七笔形态'] == FdSeven.L3B1.value:
                s['日线右侧多头因子'] = Factors.DLA1.value

            if s['日线_第N笔方向'] == Direction.Down.value and s['日线_第N笔结束标记的分型强弱'] == "强" \
                    and s['5分钟_第N笔的七笔形态'] == FdSeven.L3A1.value:
                s['日线右侧多头因子'] = Factors.DLA2.value

            if "底背弛" in s['日线_第N笔的五笔形态'] and s['5分钟_第N笔的七笔形态'] == FdSeven.L3B1.value:
                s['日线右侧多头因子'] = Factors.DLA3.value

            if "底背弛" in s['日线_第N笔的七笔形态'] and s['5分钟_第N笔的七笔形态'] == FdSeven.L3B1.value:
                s['日线右侧多头因子'] = Factors.DLA4.value

            if s['日线_第N笔方向'] == Direction.Down.value and "底背弛" in s['30分钟_第N笔的五笔形态']:
                s['日线左侧空头因子'] = Factors.DSB1.value

            if s['日线_第N笔方向'] == Direction.Down.value and "底背弛" in s['30分钟_第N笔的七笔形态']:
                s['日线左侧空头因子'] = Factors.DSB2.value

        if "日线" in self.freqs and "5分钟" in self.freqs and "1分钟" in self.freqs:
            if s['1分钟_第N笔的七笔形态'] == FdSeven.L3B1.value and s['日线_第N笔方向'] == Direction.Down.value:
                if s['5分钟_第N笔方向'] == Direction.Down.value:
                    s['5分钟右侧多头因子'] = Factors.F5LA1.value
                if "底背弛" in s['5分钟_第N笔的五笔形态']:
                    s['5分钟右侧多头因子'] = Factors.F5LA2.value
                if "底背弛" in s['5分钟_第N笔的七笔形态']:
                    s['5分钟右侧多头因子'] = Factors.F5LA3.value

            if s['1分钟_第N笔的七笔形态'] == FdSeven.S3B1.value and s['日线_第N笔方向'] == Direction.Up.value:
                if s['5分钟_第N笔方向'] == Direction.Up.value:
                    s['5分钟右侧空头因子'] = Factors.F5SA1.value
                if "顶背弛" in s['5分钟_第N笔的五笔形态']:
                    s['5分钟右侧空头因子'] = Factors.F5SA2.value
                if "顶背弛" in s['5分钟_第N笔的七笔形态']:
                    s['5分钟右侧空头因子'] = Factors.F5SA3.value
        return s

    def update_factors(self, data: List[RawBar]):
        """更新多级别联立因子"""
        for row in data:
            self.kg.update(row)

        klines_one = self.kg.get_klines({k: 1 for k in self.freqs})

        for freq, klines_ in klines_one.items():
            k = klines_[-1]
            self.kas[freq].update(k)

        self.symbol = self.kas["1分钟"].symbol
        self.end_dt = self.kas["1分钟"].bars_raw[-1].dt
        self.latest_price = self.kas["1分钟"].bars_raw[-1].close
        self.s = self._calculate_factors()




# coding: utf-8

import warnings
from collections import OrderedDict
from typing import List
from .objects import Direction, BI
from .enum import FdThree, FdFive, FdSeven, FdNine

def signals_long_short():
    """信号值的多空经验分析"""
    ls = {
        "第N笔方向": {
            Direction.Down.value: "多头信号",
            Direction.Up.value: "空头信号",
        },
        "第N笔的三笔形态": {
            FdThree.L1.value: "多头信号",
            FdThree.L2.value: "多头信号",
            FdThree.L3.value: "多头信号",
            FdThree.L4.value: "多头信号",
            FdThree.L5.value: "多头信号",
            FdThree.L6.value: "多头信号",

            FdThree.S1.value: "空头信号",
            FdThree.S2.value: "空头信号",
            FdThree.S3.value: "空头信号",
            FdThree.S4.value: "空头信号",
            FdThree.S5.value: "空头信号",
            FdThree.S6.value: "空头信号",
        },
        "第N-1笔的三笔形态": {
            FdThree.L1.value: "空头信号",
            FdThree.L2.value: "空头信号",
            FdThree.L3.value: "空头信号",
            FdThree.L4.value: "空头信号",
            FdThree.L5.value: "空头信号",
            FdThree.L6.value: "空头信号",

            FdThree.S1.value: "多头信号",
            FdThree.S2.value: "多头信号",
            FdThree.S3.value: "多头信号",
            FdThree.S4.value: "多头信号",
            FdThree.S5.value: "多头信号",
            FdThree.S6.value: "多头信号",
        },
        "第N-2笔的三笔形态": {
            FdThree.L1.value: "多头信号",
            FdThree.L2.value: "多头信号",
            FdThree.L3.value: "多头信号",
            FdThree.L4.value: "多头信号",
            FdThree.L5.value: "多头信号",
            FdThree.L6.value: "多头信号",

            FdThree.S1.value: "空头信号",
            FdThree.S2.value: "空头信号",
            FdThree.S3.value: "空头信号",
            FdThree.S4.value: "空头信号",
            FdThree.S5.value: "空头信号",
            FdThree.S6.value: "空头信号",
        },
        "第N-3笔的三笔形态": {
            FdThree.L1.value: "空头信号",
            FdThree.L2.value: "空头信号",
            FdThree.L3.value: "空头信号",
            FdThree.L4.value: "空头信号",
            FdThree.L5.value: "空头信号",
            FdThree.L6.value: "空头信号",

            FdThree.S1.value: "多头信号",
            FdThree.S2.value: "多头信号",
            FdThree.S3.value: "多头信号",
            FdThree.S4.value: "多头信号",
            FdThree.S5.value: "多头信号",
            FdThree.S6.value: "多头信号",
        },
        "第N笔的五笔形态": {
            FdFive.L1A1.value: "多头信号",
            FdFive.L1B1.value: "多头信号",
            FdFive.L2A1.value: "多头信号",
            FdFive.L2B1.value: "多头信号",
            FdFive.L3A1.value: "多头信号",
            FdFive.L4A1.value: "多头信号",
            FdFive.L4A2.value: "多头信号",
            FdFive.L4B1.value: "多头信号",
            FdFive.L4B2.value: "多头信号",
            FdFive.L4C1.value: "多头信号",
            FdFive.L4C2.value: "多头信号",
            FdFive.L4D1.value: "多头信号",
            FdFive.L4D2.value: "多头信号",
            FdFive.L5A1.value: "多头信号",
            FdFive.L5B1.value: "多头信号",

            FdFive.S1A1.value: "空头信号",
            FdFive.S1B1.value: "空头信号",
            FdFive.S2A1.value: "空头信号",
            FdFive.S2B1.value: "空头信号",
            FdFive.S3A1.value: "空头信号",
            FdFive.S4A1.value: "空头信号",
            FdFive.S4A2.value: "空头信号",
            FdFive.S4B1.value: "空头信号",
            FdFive.S4B2.value: "空头信号",
            FdFive.S4C1.value: "空头信号",
            FdFive.S4C2.value: "空头信号",
            FdFive.S4D1.value: "空头信号",
            FdFive.S4D2.value: "空头信号",
            FdFive.S5A1.value: "空头信号",
            FdFive.S5B1.value: "空头信号",
        },
        "第N-1笔的五笔形态": {
            FdFive.L1A1.value: "空头信号",
            FdFive.L1B1.value: "空头信号",
            FdFive.L2A1.value: "空头信号",
            FdFive.L2B1.value: "空头信号",
            FdFive.L3A1.value: "空头信号",
            FdFive.L4A1.value: "空头信号",
            FdFive.L4A2.value: "空头信号",
            FdFive.L4B1.value: "空头信号",
            FdFive.L4B2.value: "空头信号",
            FdFive.L4C1.value: "空头信号",
            FdFive.L4C2.value: "空头信号",
            FdFive.L4D1.value: "空头信号",
            FdFive.L4D2.value: "空头信号",
            FdFive.L5A1.value: "空头信号",
            FdFive.L5B1.value: "空头信号",

            FdFive.S1A1.value: "多头信号",
            FdFive.S1B1.value: "多头信号",
            FdFive.S2A1.value: "多头信号",
            FdFive.S2B1.value: "多头信号",
            FdFive.S3A1.value: "多头信号",
            FdFive.S4A1.value: "多头信号",
            FdFive.S4A2.value: "多头信号",
            FdFive.S4B1.value: "多头信号",
            FdFive.S4B2.value: "多头信号",
            FdFive.S4C1.value: "多头信号",
            FdFive.S4C2.value: "多头信号",
            FdFive.S4D1.value: "多头信号",
            FdFive.S4D2.value: "多头信号",
            FdFive.S5A1.value: "多头信号",
            FdFive.S5B1.value: "多头信号",
        },
        "第N-2笔的五笔形态": {
            FdFive.L1A1.value: "多头信号",
            FdFive.L1B1.value: "多头信号",
            FdFive.L2A1.value: "多头信号",
            FdFive.L2B1.value: "多头信号",
            FdFive.L3A1.value: "多头信号",
            FdFive.L4A1.value: "多头信号",
            FdFive.L4A2.value: "多头信号",
            FdFive.L4B1.value: "多头信号",
            FdFive.L4B2.value: "多头信号",
            FdFive.L4C1.value: "多头信号",
            FdFive.L4C2.value: "多头信号",
            FdFive.L4D1.value: "多头信号",
            FdFive.L4D2.value: "多头信号",
            FdFive.L5A1.value: "多头信号",
            FdFive.L5B1.value: "多头信号",

            FdFive.S1A1.value: "空头信号",
            FdFive.S1B1.value: "空头信号",
            FdFive.S2A1.value: "空头信号",
            FdFive.S2B1.value: "空头信号",
            FdFive.S3A1.value: "空头信号",
            FdFive.S4A1.value: "空头信号",
            FdFive.S4A2.value: "空头信号",
            FdFive.S4B1.value: "空头信号",
            FdFive.S4B2.value: "空头信号",
            FdFive.S4C1.value: "空头信号",
            FdFive.S4C2.value: "空头信号",
            FdFive.S4D1.value: "空头信号",
            FdFive.S4D2.value: "空头信号",
            FdFive.S5A1.value: "空头信号",
            FdFive.S5B1.value: "空头信号",
        },
        "第N-3笔的五笔形态": {
            FdFive.L1A1.value: "空头信号",
            FdFive.L1B1.value: "空头信号",
            FdFive.L2A1.value: "空头信号",
            FdFive.L2B1.value: "空头信号",
            FdFive.L3A1.value: "空头信号",
            FdFive.L4A1.value: "空头信号",
            FdFive.L4A2.value: "空头信号",
            FdFive.L4B1.value: "空头信号",
            FdFive.L4B2.value: "空头信号",
            FdFive.L4C1.value: "空头信号",
            FdFive.L4C2.value: "空头信号",
            FdFive.L4D1.value: "空头信号",
            FdFive.L4D2.value: "空头信号",
            FdFive.L5A1.value: "空头信号",
            FdFive.L5B1.value: "空头信号",

            FdFive.S1A1.value: "多头信号",
            FdFive.S1B1.value: "多头信号",
            FdFive.S2A1.value: "多头信号",
            FdFive.S2B1.value: "多头信号",
            FdFive.S3A1.value: "多头信号",
            FdFive.S4A1.value: "多头信号",
            FdFive.S4A2.value: "多头信号",
            FdFive.S4B1.value: "多头信号",
            FdFive.S4B2.value: "多头信号",
            FdFive.S4C1.value: "多头信号",
            FdFive.S4C2.value: "多头信号",
            FdFive.S4D1.value: "多头信号",
            FdFive.S4D2.value: "多头信号",
            FdFive.S5A1.value: "多头信号",
            FdFive.S5B1.value: "多头信号",
        },
        "第N笔的七笔形态": {
            FdSeven.L1A1.value: "多头信号",
            FdSeven.L2A1.value: "多头信号",
            FdSeven.L3A1.value: "多头信号",
            FdSeven.L3B1.value: "多头信号",
            FdSeven.L4A1.value: "多头信号",

            FdSeven.S1A1.value: "空头信号",
            FdSeven.S2A1.value: "空头信号",
            FdSeven.S3A1.value: "空头信号",
            FdSeven.S3B1.value: "空头信号",
            FdSeven.S4A1.value: "空头信号",
        },
        "第N-1笔的七笔形态": {
            FdSeven.L1A1.value: "空头信号",
            FdSeven.L2A1.value: "空头信号",
            FdSeven.L3A1.value: "空头信号",
            FdSeven.L3B1.value: "空头信号",
            FdSeven.L4A1.value: "空头信号",

            FdSeven.S1A1.value: "多头信号",
            FdSeven.S2A1.value: "多头信号",
            FdSeven.S3A1.value: "多头信号",
            FdSeven.S3B1.value: "多头信号",
            FdSeven.S4A1.value: "多头信号",
        },
        "第N-2笔的七笔形态": {
            FdSeven.L1A1.value: "多头信号",
            FdSeven.L2A1.value: "多头信号",
            FdSeven.L3A1.value: "多头信号",
            FdSeven.L3B1.value: "多头信号",
            FdSeven.L4A1.value: "多头信号",

            FdSeven.S1A1.value: "空头信号",
            FdSeven.S2A1.value: "空头信号",
            FdSeven.S3A1.value: "空头信号",
            FdSeven.S3B1.value: "空头信号",
            FdSeven.S4A1.value: "空头信号",
        },
        "第N-3笔的七笔形态": {
            FdSeven.L1A1.value: "空头信号",
            FdSeven.L2A1.value: "空头信号",
            FdSeven.L3A1.value: "空头信号",
            FdSeven.L3B1.value: "空头信号",
            FdSeven.L4A1.value: "空头信号",

            FdSeven.S1A1.value: "多头信号",
            FdSeven.S2A1.value: "多头信号",
            FdSeven.S3A1.value: "多头信号",
            FdSeven.S3B1.value: "多头信号",
            FdSeven.S4A1.value: "多头信号",
        },
        "第N笔的九笔形态": {
            FdNine.L1A1.value: "多头信号",
            FdNine.L2A1.value: "多头信号",
            FdNine.L2B1.value: "多头信号",
            FdNine.L2C1.value: "多头信号",
            FdNine.L3A1.value: "多头信号",
            FdNine.L4A1.value: "多头信号",
            FdNine.L4B1.value: "多头信号",

            FdNine.S1A1.value: "空头信号",
            FdNine.S2A1.value: "空头信号",
            FdNine.S2B1.value: "空头信号",
            FdNine.S2C1.value: "空头信号",
            FdNine.S3A1.value: "空头信号",
            FdNine.S4A1.value: "空头信号",
            FdNine.S4B1.value: "空头信号",
        },
        "第N-1笔的九笔形态": {
            FdNine.L1A1.value: "空头信号",
            FdNine.L2A1.value: "空头信号",
            FdNine.L2B1.value: "空头信号",
            FdNine.L2C1.value: "空头信号",
            FdNine.L3A1.value: "空头信号",
            FdNine.L4A1.value: "空头信号",
            FdNine.L4B1.value: "空头信号",

            FdNine.S1A1.value: "多头信号",
            FdNine.S2A1.value: "多头信号",
            FdNine.S2B1.value: "多头信号",
            FdNine.S2C1.value: "多头信号",
            FdNine.S3A1.value: "多头信号",
            FdNine.S4A1.value: "多头信号",
            FdNine.S4B1.value: "多头信号",
        },
        "第N-2笔的九笔形态": {
            FdNine.L1A1.value: "多头信号",
            FdNine.L2A1.value: "多头信号",
            FdNine.L2B1.value: "多头信号",
            FdNine.L2C1.value: "多头信号",
            FdNine.L3A1.value: "多头信号",
            FdNine.L4A1.value: "多头信号",
            FdNine.L4B1.value: "多头信号",

            FdNine.S1A1.value: "空头信号",
            FdNine.S2A1.value: "空头信号",
            FdNine.S2B1.value: "空头信号",
            FdNine.S2C1.value: "空头信号",
            FdNine.S3A1.value: "空头信号",
            FdNine.S4A1.value: "空头信号",
            FdNine.S4B1.value: "空头信号",
        },
        "第N-3笔的九笔形态": {
            FdNine.L1A1.value: "空头信号",
            FdNine.L2A1.value: "空头信号",
            FdNine.L2B1.value: "空头信号",
            FdNine.L2C1.value: "空头信号",
            FdNine.L3A1.value: "空头信号",
            FdNine.L4A1.value: "空头信号",
            FdNine.L4B1.value: "空头信号",

            FdNine.S1A1.value: "多头信号",
            FdNine.S2A1.value: "多头信号",
            FdNine.S2B1.value: "多头信号",
            FdNine.S2C1.value: "多头信号",
            FdNine.S3A1.value: "多头信号",
            FdNine.S4A1.value: "多头信号",
            FdNine.S4B1.value: "多头信号",
        },
    }

    ls_map = OrderedDict()
    for freq in ['1分钟', '5分钟', '15分钟', '30分钟', '日线']:
        for name, values in ls.items():
            for v, s in values.items():
                k = "{}_{}_{}".format(freq, name, v)
                ls_map[k] = s
    return ls_map

def check_three_fd(fds: List[BI]) -> str:
    """识别三段形态

    :param fds: list
        由远及近的三段形态
    :return: str
    """
    v = FdThree.Other.value

    if len(fds) != 3:
        warnings.warn("len(fdx) != 3，无法识别三段形态")
        return v

    fd1, fd2, fd3 = fds
    if not (fd1.direction == fd3.direction):
        # warnings.warn("1,3的 direction 不一致，无法识别三段形态")
        print("1,3的 direction 不一致，无法识别三段形态")
        return v

    if fd3.direction == Direction.Down:
        if fd3.low > fd1.high:
            v = FdThree.L1.value        # "向下不重合"

        if fd2.low < fd3.low < fd1.high < fd2.high:
            v = FdThree.L2.value        # "向下奔走型中枢"

        if fd1.high > fd3.high and fd1.low < fd3.low:
            v = FdThree.L3.value        # "向下三角收敛中枢"

        if fd1.high < fd3.high and fd1.low > fd3.low:
            v = FdThree.L4.value        # "向下三角扩张中枢"

        if fd3.low < fd1.low and fd3.high < fd1.high:
            if fd3.power < fd1.power:
                v = FdThree.L5.value    # '向下盘背中枢'
            else:
                v = FdThree.L6.value    # '向下无背中枢'
    elif fd3.direction == Direction.Up:
        if fd3.high > fd1.low:
            v = FdThree.S1.value        # "向上不重合"

        if fd2.low < fd1.low < fd3.high < fd2.high:
            v = FdThree.S2.value        # "向上奔走型中枢"

        if fd1.high > fd3.high and fd1.low < fd3.low:
            v = FdThree.S3.value        # "向上三角收敛中枢"

        if fd1.high < fd3.high and fd1.low > fd3.low:
            v = FdThree.S4.value        # "向上三角扩张中枢"

        if fd3.low > fd1.low and fd3.high > fd1.high:
            if fd3.power < fd1.power:
                v = FdThree.S5.value        # '向上盘背中枢'
            else:
                v = FdThree.S6.value        # '向上无背中枢'
    else:
        raise ValueError("direction 的取值错误")

    return v

def check_five_fd(fds: List[BI]) -> str:
    """识别五段形态

    :param fds: list
        由远及近的五段走势
    :return: str
    """
    v = FdFive.Other.value

    if len(fds) != 5:
        warnings.warn("len(fdx) != 5，无法识别五段形态")
        return v

    fd1, fd2, fd3, fd4, fd5 = fds
    if not (fd1.direction == fd3.direction == fd5.direction):
        warnings.warn("1,3,5 的 direction 不一致，无法识别五段形态")
        return v
    direction = fd1.direction
    max_high = max([x.high for x in fds])
    min_low = min([x.low for x in fds])

    if direction == Direction.Down:
        if fd1.high > fd3.high > fd5.high and fd1.low < fd3.low < fd5.low:
            v = FdFive.L1A1.value     # "向下三角收敛中枢"
        if fd1.high < fd3.high < fd5.high and fd1.low > fd3.low > fd5.low:
            v = FdFive.L1B1.value     # "向下三角扩张中枢"

        if max_high == fd1.high and min_low == fd5.low and fd4.high > fd2.low and fd5.power < fd1.power:
            if fd2.high > fd4.high and fd2.low < fd4.low:
                v = FdFive.L2B1.value     # "aAb式底背驰B"
            else:
                v = FdFive.L2A1.value     # "aAb式底背驰A"

        if max_high == fd1.high and min_low == fd3.low and fd5.power < fd1.power \
                and fd4.high > fd2.low and fd5.low > fd3.low:
            v = FdFive.L2C1.value     # "aAb式底背驰C"

        if fd4.high < fd2.low and fd5.power < fd3.power and max_high == fd1.high and min_low == fd5.low:
            v = FdFive.L3A1.value     # "类趋势底背驰"

    elif direction == Direction.Up:
        if fd1.high > fd3.high > fd5.high and fd1.low < fd3.low < fd5.low:
            v = FdFive.S1A1.value     # "向上三角收敛中枢"
        if fd1.high < fd3.high < fd5.high and fd1.low > fd3.low > fd5.low:
            v = FdFive.S1B1.value     # "向上三角扩张中枢"

        if max_high == fd5.high and min_low == fd1.low and fd5.power < fd1.power:
            if fd2.high > fd4.high and fd2.low < fd4.low:
                v = FdFive.S2B1.value     # "aAb式顶背驰B"
            else:
                v = FdFive.S2A1.value     # "aAb式顶背驰A"

        if max_high == fd3.high and min_low == fd1.low and fd5.power < fd1.power \
                and fd5.high < fd3.high and fd4.low < fd2.high:
            v = FdFive.S2C1.value     # "aAb式顶背驰C"

        if max_high == fd5.high and min_low == fd1.low and fd5.power < fd1.power and fd4.low > fd4.low:
            v = FdFive.S3A1.value     # "类趋势顶背驰"
    else:
        raise ValueError("direction 的取值错误，必须是 down / up 之一，实际值为{}".format(fd5.direction))

    direction = fd5.direction
    zg = min([x.high for x in fds[:3]])
    gg = max([x.high for x in fds[:3]])
    zd = max([x.low for x in fds[:3]])
    dd = min([x.low for x in fds[:3]])
    max_high = max([x.high for x in fds])
    min_low = min([x.low for x in fds])

    if zd > zg:
        return v

    if direction == Direction.Down:
        if fd4.high == max_high:
            if fd1.high > fd3.high and fd1.low < fd3.low:
                if fd5.low > gg:
                    v = FdFive.L4A1.value     # "三买A1"
                if gg > fd5.low > zg:
                    v = FdFive.L4A2.value     # "三买A2"

            if fd3.high > fd1.high and fd3.low < fd1.low:
                if fd5.low > gg:
                    v = FdFive.L4B1.value     # "三买B1"
                if gg > fd5.low > zg:
                    v = FdFive.L4B2.value     # "三买B2"

            if fd3.high > fd1.high and fd3.low > fd1.low:
                if fd5.low > gg:
                    v = FdFive.L4C1.value     # "三买C1"
                if gg > fd5.low > zg:
                    v = FdFive.L4C2.value     # "三买C2"

            if fd3.power < fd1.power and fd3.high < fd1.high and fd3.low < fd1.low:
                if fd5.low > gg:
                    v = FdFive.L4D1.value     # "三买D1"
                if gg > fd5.low > zg:
                    v = FdFive.L4D2.value     # "三买D2"

    elif direction == Direction.Up:
        if fd4.low == min_low:
            if fd1.high > fd3.high and fd1.low < fd3.low:
                if fd5.high < dd:
                    v = FdFive.S4A1.value     # "三卖A1"
                if zd > fd5.high > dd:
                    v = FdFive.S4A2.value     # "三卖A2"

            if fd3.high > fd1.high and fd3.low < fd1.low:
                if fd5.high < dd:
                    v = FdFive.S4B1.value     # "三卖B1"
                if zd > fd5.high > dd:
                    v = FdFive.S4B2.value     # "三卖B2"

            if fd3.high < fd1.high and fd3.low < fd1.low:
                if fd5.high < dd:
                    v = FdFive.S4C1.value     # "三卖C1"
                if zd > fd5.high > dd:
                    v = FdFive.S4C2.value     # "三卖C2"

            if fd3.high > fd1.high and fd3.low > fd1.low and fd3.power < fd1.power:
                if fd5.high < dd:
                    v = FdFive.S4D1.value     # "三卖D1"
                if zd > fd5.high > dd:
                    v = FdFive.S4D2.value     # "三卖D2"
    else:
        raise ValueError("direction 的取值错误")
    return v

def check_seven_fd(fds: List[BI]) -> str:
    """识别七段形态

    :param fds: list
        由远及近的七段走势
    :return: str
    """
    v = FdSeven.Other.value
    if len(fds) != 7:
        warnings.warn("len(fdx) != 7，无法识别七段形态")
        return v

    fd1, fd2, fd3, fd4, fd5, fd6, fd7 = fds
    max_high = max([x.high for x in fds])
    min_low = min([x.low for x in fds])

    if fd7.direction == Direction.Down:
        if fd1.high == max_high and fd7.low == min_low:
            if fd2.low > fd4.high and fd4.low > fd6.high and fd7.power < fd5.power:
                v = FdSeven.L4A1.value    # 七笔类趋势底背驰

            if fd2.low > fd4.high and fd6.high < min(fd2.low, fd4.low) and fd7.power < fd1.power:
                v = FdSeven.L1A1.value  # "aAb式底背驰"

            if fd2.low > fd4.high and fd4.low > fd6.high and fd7.power < fd5.power:
                v = FdSeven.L2A1.value  # "aAbcd式底背驰"

        if min_low == fd3.low and fd1.high > fd3.high and max_high == max(fd1.high, fd4.high) \
                and fd5.high > fd7.high and fd5.low < fd7.low:
            v = FdSeven.L3A1.value  # "BaA式右侧底A"

        if min_low == fd3.low and fd2.low > fd4.low < fd6.low \
                and fd2.high > fd4.high < fd6.high and fd7.low > fd3.low and fd4.high > fd2.low:
            v = FdSeven.L3B1.value  # "BaA式右侧底B"

    elif fd7.direction == Direction.Up:
        if fd1.low == min_low and fd7.high == max_high:
            if fd2.high < fd4.low and fd4.high < fd6.low and fd7.power < fd5.power:
                v = FdSeven.S4A1.value  # "类趋势顶背驰"

            if fd4.low < fd2.high and fd6.low < fd2.high and fd7.power < fd1.power:
                v = FdSeven.S1A1.value  # "aAb式顶背驰"

            if fd4.low < fd2.high and fd6.low > max(fd4.high, fd2.high) and fd7.power < fd5.power:
                v = FdSeven.S2A1.value  # "aAbcd式顶背驰"

        if max_high == fd4.high and min_low == min(fd4.low, fd1.low) and fd6.high < fd2.low \
                and fd5.high > fd7.high and fd5.low < fd7.low and fd1.low < fd3.low:
            v = FdSeven.S3A1.value  # "BaA式右侧顶A"

        if max_high == fd4.high and fd2.high < fd4.high > fd6.high \
                and fd2.low < fd4.low > fd6.low and fd7.high < fd3.high and fd4.low < fd2.high:
            v = FdSeven.S3B1.value  # "BaA式右侧顶B"

    else:
        raise ValueError("direction 的取值错误")
    return v

def check_nine_fd(fds: List[BI]) -> str:
    """识别九段形态

    :param fds: list
        由远及近的九段形态
    :return: str
    """
    v = FdNine.Other.value
    if len(fds) != 9:
        warnings.warn("len(fdx) != 9，无法识别九段形态")
        return v

    direction = fds[-1].direction
    fd1, fd2, fd3, fd4, fd5, fd6, fd7, fd8, fd9 = fds
    max_high = max([x.high for x in fds])
    min_low = min([x.low for x in fds])

    if direction == Direction.Up:
        if min_low == fd1.low and max_high == fd9.high:
            if fd2.high < fd4.low and fd4.high < fd6.low and fd6.high < fd8.low and fd9.power < fd7.power:
                v = FdNine.S1A1.value   # 类趋势顶背驰

            if fd2.high < fd4.low and fd4.high < fd6.low and fd6.high > fd8.low:
                v = FdNine.S2A1.value   # 含有一个下上下方向中枢A

            if fd2.high < fd4.low and fd4.high > fd6.low and fd6.high < fd8.low:
                v = FdNine.S2B1.value   # 含有一个下上下方向中枢B

            if fd2.high > fd4.low and fd4.high < fd6.low and fd6.high < fd8.low:
                v = FdNine.S2C1.value   # 含有一个下上下方向中枢C

            if fd2.high > fd4.low and fd4.high < fd6.low \
                    and fd6.high > fd8.low > max(fd2.high, fd4.high)\
                    and fd6.low > max(fd2.high, fd4.high):
                v = FdNine.S3A1.value   # 标准aAbBc向上趋势

            if fd2.high < fd4.low and fd4.high > fd6.low and fd4.high > fd8.low:
                v = FdNine.S4A1.value   # 上中枢七段上涨

            if fd2.high > fd4.low and fd2.high > fd6.low and fd6.high < fd8.low:
                v = FdNine.S4B1.value   # 下中枢七段上涨
    elif direction == Direction.Down:
        if max_high == fd1.high and min_low == fd9.low:
            if fd2.low > fd4.high and fd4.low > fd6.high and fd6.low > fd8.high and fd9.power < fd7.power:
                v = FdNine.L1A1.value   # 类趋势底背驰

            if fd2.low > fd4.high and fd4.low > fd6.high and fd6.low < fd8.high:
                v = FdNine.L2A1.value   # 含有一个上下上方向中枢A
            if fd2.low > fd4.high and fd4.low < fd6.high and fd6.low > fd8.high:
                v = FdNine.L2B1.value   # 含有一个上下上方向中枢B
            if fd2.low < fd4.high and fd4.low > fd6.high and fd6.low > fd8.high:
                v = FdNine.L2C1.value   # 含有一个上下上方向中枢C

            if fd2.low < fd4.high and fd4.low > fd6.high \
                    and fd6.low < fd8.high < min(fd2.low, fd4.low)\
                    and fd6.high < min(fd2.low, fd4.low):
                v = FdNine.L3A1.value   # 标准aAbBc向下趋势

            if fd2.low < fd4.high and fd2.low < fd6.high and fd6.low > fd8.high:
                v = FdNine.L4A1.value   # 上中枢七段下跌
            if fd2.low > fd4.high and fd4.low < fd6.high and fd4.low < fd8.high:
                v = FdNine.L4B1.value   # 下中枢七段下跌
    else:
        raise ValueError("direction 的取值错误")
    return v
